////////////////////////////////////////////////////////////////////////////////
//   Server WinShoeSMTP
//
//
////////////////////////////////////////////////////////////////////////////////
//  Based on RFC 821 - The SMTP Protocol
// ..
// Author Ray Malone
// Version .10 Started July 10, 1999 Completd July 30, 1999
//
// the WinshoeSMPT server is based on TWinshoeSMTPListener  which is an abstract
// component that just listens on the port and parses the received commands. It
// calls abstract procedures for each command; Events for each of these
// comands are defined in the listener.
// The TWinshoeSPTPServer decends from the TWinshoeSMTPListener and implements
// The commands and events.
// ON receipt of a commmand the DoCommandXXXX procedure is called and if assigned
// it calls the related event. if the event is not assigned then the default
// behavior is run.
// Each event has the boolean variable named Handled. if Handled is set to True
// in the event handler default processing is not done.
//
//------------------------------------------------------------------------------
// The simplest Mail transaction is as follows
// An SMTP Client makes a connection.
//
// The Doconnect sends the following reply
//      220 '+ComputerName +' Simple Mail Transfer Service Ready
// The Client then sends a HELO command Which consists of the HELO command
//  and the Clients Domain Name
// The DoCommandHelo then replies
//        '250 '+ SMTPDomainName
//  Where ComputerName is the name returned by Win function GetComputerName
//
//  The Client then sends the MAIL command which contains the From address
//    For example: MAIL From: "Charlie Smith" <char@hisdomain.com>
//        or     MAIL From: char@hisdomain.com
//   The Mail comannd argument is stored in  fFromStr which is th FromStr property
//   if the address part is valid ( that is chars, @ chars dot chars) the server
//   Replies 250 OK
// The Client then sends one or more RCPT commans in any of the following forms:
//        RCPT To: "Bill Jones" <bill@hisdomain.com> (this is a comment)
//        RCPT To: "Bill Jones" <bill@hisdomain.com>
//        RCPT To: <bill@hisdomain.com>
//        RCPT To: bill@hisdomain.com
//  if the addess part is not valid( that is chars, @ chars dot chars) the server
//  replies 501 Syntax error
//  if the address part is valid it checks to see if the domain is local. if not
//  the reply is 250 (should be 251) that it will forward teh mail.
//  if the domain is local a check is made to see if the mail box (User Id) is
//  valid. if it is not then a 550 ERR invalid mail box error is returned.

//   The RCPT comannd arguments is stored in  fRecipients tRecipientList
//  this list contains the Addrsss part of the RCPT command as the Addresss
//  Field. A Local Mail boolean, and a Local client boolean. if both the
//  LocalMail and LocalClient are true, the path to the recipients mailbox is
//  stored in the MailBoxDir field.
//  Mail to Be forwared only has the address field set.

// The Client next sends the data command.
//   The server responds 354 Start Mail Input; end with <CRLF>.<CRLF>
// The client then sends text lines of the Email document which are stored in
//   the FData.StringList
//  When the <CRLF>.<CRLF> is received by the SErver it then stores the
//  DoCommandData then Calls WriteMailToUserDirOrForward whick drops local mail
//  to the Users MailBox or writes it to the  Pickup dir to be forwarded.
//
//  when it is finished.
//  DoCommandData then Sends the folowing reply
//   250 OK
//  At this point the client can call quit or start another Mail command
//-----------------------------------------------------------------------------
// Revision History
// 10/15/99
// 0.01 Modification to handle  extra Period and blank line
// Supplied by Roman Evstyunichev... thanks.
//
// 10/15/99
// 0.01 Revised Writetodisk to make sure To: is the intended Recipient for To
// CC and BC do not depend on x-receiver
////////////////////////////////////////////////////////////////////////////////
{
v0.11  Added logging to the IMS INLOG directory, same format as they use.
       (554 RBL rejections should show up here too)
v0.12  Fixed bug in the localnet masking routine that allowed most IP
       addresses to appear local
v0.13-0.15
       Fixed a difference between the way SMTPRS and Winshoes deals with
       the terminating period that stops the DATA transfer.  (This caused
       no non-local mail to get delivered).
v0.16  Fixed address parsing engine to correctly deal with !, % and :
       This then correctly parses and address to be a relay hack, and
       allows the server to reject the address.
v0.17  Corrected startup bug that caused service to crash when started on
       some NT4 workstations. (in mainline)
v0.18  Made MAIL FROM: <> reply RFC 2505 compatible (server now no longer
       cares what's there and responds with a 250 OK regardless).
v0.19  When IP fails to rDNS, the header says unverified instead of a blank
       spot.
v0.20  When local IP fails to rDNS, the server name changes to an IP
       instead of blank, fixing message-id problems
v0.21  Deleted an unneccesary check for localdomain
v0.22  Trying to gain some speed on the write for large files
v0.23  OK, that didn't work :)
v0.24  Removed date insertion when another received line is found
       to track down a possible bug where Netscape and Outlook puke when
       downloading mail.
v0.25  Fixed bonehead Date bug.
v0.26  3 different ways to kill an RBL listed server connection now:
       1) immediate termination via a 421 command
       2) any rcpt to: and data command result in a 554
       3) first rcpt to: results in a 554 and a killed connection
       Also fixed so that no valid RCPT results in a 503 No valid recipients specified
v0.27  Control values now read from our own registry settings.  A
       control panel app is included.
v0.28  By some miracle I discovered that if two threads hit the server at the
       exact same time, the RBL info can pass from one thread to the other.
v0.29  Fix in 0.28 removed access denied log file entries, added back in.
v0.30  Fix in 0.29 caused abstract method crash.
v0.31  After a DATA command finishes, flush the message.  Apparently some
       mailing list servers don't issue an RSET before starting the next message.
v0.32  Trying to make servername resolve correctly.
v0.33  Added localuser deny (notify MTU when user doesn't exist).  Pulled from
       the alias listing for IMS. (probably won't work for list aliases)
v0.34  Fixed time offset notation by adding calc for daylight savings
v0.35  Added TCriticalSections around log file reads/writes to make multithread
       safe.  Also added optional error logging to the inlog which adds errors
       for invalid local names and relay attempts.  Made localuser deny (0.33)
       list compatible.
v0.36  Added capability to use plugin1's config file to extract user list and
       ability to check NT user database for user name (Thanks Klint Gore).
       Speed seems to be an issue with the username lookup.
v0.37  added list-request for EMWAC mailing lists.
v0.38  adding "nobody@a.com" to the plugin1 config file causes all mail
       being delivered for a.com to be accepted without passing the user check
v0.39  fixed a small problem with plugin1 file parsing and corrected enumerated
       string order for command list
v0.40  All date time strings now in english format
v0.41  Changed subnet internal structure to add local black and white list.
       Eliminates the need for a local DNS server
v0.42  NT User verification changed to reading all members of a group at
       service start
v0.43  Rebuild with fix in NT User Verification
v0.44  After much wrestling with the user verification code, I finally
       got all the bugs out of it.  The API can be such a pain in the ass.
       Also added -version as a command line parameter.
v0.45  If a socket dumps on a socket error (remote terminates for example),
       the message was not getting disposed, causing a large memory leak.
       put try..finally block around socket code to handle this problem.
v0.46  Removed a mountain of unneccessary code in effort to witchhunt
       mysterious crashes
v0.47  Added spamtrapping.  You can specify a specific list of addresses
       which are not valid, and the system will cut off that IP address
       for a fixed period of time.
}

unit ServerWinshoeSMTP;

interface

uses
  Classes,Dialogs,SyncObjs,
  Winshoes,WinShoeMessage,ServerWinshoe,
  SUBNET;

const
  VersionString = '0.47';
  ProductName = 'SMTPRCV';
  MinutesPerDay = 24*60;

type
  TSMTPCommands =
  (cError,cHELO,cMAIL,cRCPT,cDATA,cSEND,cSOML,cSAML,
   cRSET,CVRFY,cEXPN,cHELP,cNOOP,cQuit,cTURN);

const
  FirstSMTPCommand = cHELO;
  LastSMTPCommand = cTURN;

  SMTPCommands : array [TSMTPCommands] of string =
   ('Error','HELO','MAIL','RCPT','DATA','SEND','SOML','SAML',
    'RSET','VRFY','EXPN','HELP','NOOP','QUIT','TURN');



type
  pRecipientRec = ^tRecipientRec;
  tRecipientRec = record
    LocalDomain : boolean;
    ValidatedAddress : string;
  end;

type
  tRecipientList = class(TComponent)
  private
    fList : TList;
    function GetCount : integer;
  protected
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy;  override;
  published
    property  Count : integer read GetCount;
    procedure Add(RecipientItem : tRecipientRec);
    procedure Clear;
    function  GetRecipientRec(Idx: integer):tRecipientRec;
  end;


type
  TSMTPMsg = class(TComponent)
    private
      fSendingIP : string;
      fSendingDomain : string; {HELO}
      fVerifiedSendingDomain : string; {rDNS}
      fFromStr : string;
      fRecipients : TRecipientList;
      fData : TStringList;
      fDateTime : TDateTime;
      fMsgId : string;
      {vvvv 0.28}
      fFailReason : string;       {0.26}
      fDenyAllCommands : boolean; {0.26}
      fLocalSender : boolean;
    protected
    public
      constructor Create(AOwner: TComponent); override;
      destructor Destroy; override;
      procedure Clear(FirstTime : boolean);
    published
      property SendingDomain : string read fSendingDomain write fSendingDomain;
      property VerifiedSendingDomain : string read fVerifiedSendingDomain write fVerifiedSendingDomain;
      property SendingIP : string read fSendingIP write fSendingIP;
      property Recipients : TRecipientList read fRecipients write fRecipients;
      property FromStr : string read fFromStr write fFromStr;
      property Data : TStringList read fdata write fdata;
      property DateTime : TDateTime read FDateTime write FdateTime;
      property MsgId : string read FmsgId write fMsgId;
    end;

type
  PDNSListRec = ^TDNSListRec;
  TDNSListRec = record
    DNSZoneFileName : shortstring;
    DNSReponseString : shortstring;
  end;

  PSpamTrapRec = ^TSpamTrapRec;
  TSpamTrapRec = record
    CreateTime : TDateTime;
    Address : longword;
  end;

  tAccessControl = class(TComponent)
  private
    fLocalNets : TList; {of PSubNet}
    fLocalBlackList : TList; {of PSubNet}
    fWhiteList : TList; {of PSubNet}
    fDNSBlocklists : TList; {of PDNSListRec}
    fSpamTrapRelayList : TList; {of PSpamTrapRec}
    fLocalDomains : TStringList;
    fLocalUsers : TStringList;
    fSpamTrapUsers : TStringList;
    fSkipUserVerifyDomains : TStringList;
    LocalDomainsCritSection : TCriticalSection;
    LocalUsersCritSection : TCriticalSection;
    SpamTrapUsersCritSection : TCriticalSection;
    SkipCritSection : TCriticalSection;
    LocalNetCritSection : TCriticalSection;
    LocalBlackListCritSection : TCriticalSection;
    WhiteListCritSection : TCriticalSection;
    DNSListCritSection : TCriticalSection;
    SpamTrapListCritSection : TCriticalSection;
    function IsInNetList(const IP : string; fNetList : TList): boolean;
  protected
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy;  override;
  published
    function IsLocalNet(const IP : string) : boolean;
    function IsWhiteListed(const IP : string) : boolean;
    function IsBlackListedLocally(const IP : string) : boolean;
    function IsDNSBlockListed(const IP : string; var Reason : string) : boolean;
    function IsSpamTrapRelay(const IP : string; TimeOut : longint) : boolean;
    function IsLocalDomain(const Domain : string): boolean;
    function IsLocalUser(const Name : string): boolean; {0.33}
    function IsSpamTrapUser(const Name : string): boolean; {0.47}
    procedure RegisterSpamRelay(const IP : string);
    procedure AddLocalDomain(const Name : string);
    procedure AddLocalNet(LocalNet : PSubNet);
    procedure AddToLocalBlackList(BlackListEntry : PSubNet);
    procedure AddToWhiteList(WhiteListEntry : PSubNet);
    procedure AddLocalUser(const Name : string);
    procedure AddSpamTrapUser(const Name : string);
    procedure AddDomainToSkipUserVerify(const DomainToSkip : string);
    procedure AddDNSBlockList(BlockListEntry : PDNSListRec);
  end;

type
  TNextIDEvent= procedure(var ID: integer; var Handled : boolean) of object;

  TPOPAuthEvent = procedure(const IPAddy : string; var Valid : boolean;
                            var Handled : boolean) of object;
  TLogFileEvent = procedure(const LogFileEntry : string) of object;{0.11}

  TWinshoeSMTPServer = class(TWinshoeListener)
    private
      fMailWriteDir : shortstring;
      fServerName : shortstring;
      fLocalBlackListNotice : shortstring;

      fDoPOPAuth: boolean;
      fDoVerifyDomain : boolean; {replaces unverified with rDNS}
      fForwardMail : boolean; {relay all mail}
      fCheckUser : boolean; {see if user exists}
      fExtendedErrorRecording : boolean; {record user errors and relay attempts}
      fDoDNSLists : boolean;
      fDoSpamtrapping : boolean;

      fAccessControl : tAccessControl;

      fOnNextIDEvent : TNextIDEvent;
      fonPOPAuth: TPOPAuthEvent;
      fOnLogFileEvent : TLogFileEvent;{0.11}{0.29}

      fSpamTrapRejectTime : longint;
    protected
      function GetLocalFQDNName : string;
      function DoPOPLogVerify(SendingIP : string) : boolean;
      function VerifyLocalSender(SendingIP : string) : boolean;
      function PassedBlackLists(SendingIP : string; var Failure : string) : boolean;
      function VerifySendingDomain(SendingIP : string) : string;
      procedure DoConnect(Thread: TWinshoeServerThread); override;
      function  DoExecute(Thread: TWinshoeServerThread): boolean; override;
      procedure DoCommandSOML(Thread: TWinshoeServerThread; FromStr: string;
                              SessionMsg: tSMTPMsg);
      procedure DoCommandSAML(Thread: TWinshoeServerThread; FromStr: string;
                              SessionMsg: tSMTPMsg);
      procedure DoCommandVRFY(Thread: TWinshoeServerThread; aName: string);
      procedure DoCommandEXPN(Thread: TWinshoeServerThread; aListName: string);
      procedure DoCommandHELP(Thread: TWinshoeServerThread; aCommand: string);
      procedure DoCommandNOOP(Thread: TWinshoeServerThread);
      procedure DoCommandTURN(Thread: TWinshoeServerThread; SessionMsg: tSMTPMsg);
      procedure DoCommandSEND(Thread: TWinshoeServerThread; FromStr: string;
                              SessionMsg: tSMTPMsg);
      procedure DoCommandError(Thread: TWinshoeServerThread; ArgSTr: string);
      procedure DoCommandHELO(Thread: TWinshoeServerThread; Domain: string;
                              SessionMsg: tSMTPMsg);
      procedure DoCommandMAIL(Thread: TWinshoeServerThread; FromStr: string;
                              SessionMsg: tSMTPMsg);
      procedure DoCommandRCPT(Thread: TWinshoeServerThread; ToStr: string;
                              SessionMsg: tSMTPMsg);
      procedure DoCommandDATA(Thread: TWinshoeServerThread; SessionMsg: tSMTPMsg);
      procedure DoCommandQUIT(Thread: TWinshoeServerThread);
      procedure DoCommandRSET(Thread: TWinshoeServerThread; SessionMsg: tSMTPMsg);
      function GetNextID : longint;
      function WriteMessageToFile(SessionMsg: TSMTPMsg): boolean;
      function WriteMail(Thread : TWinshoeServerThread; SessionMsg: tSMTPMsg): boolean;
    public
      constructor Create(AOwner : TComponent); override;
      destructor Destroy; override;
    published
      property OnNextIDEvent : TNextIDEvent read fOnNextIDEvent write fOnNextIDEvent;
      property OnPopAuth : TPOPAuthEvent read fOnPOPAuth write fOnPOPAuth;
      property OnLogFileEvent : TLogFileEvent read fOnLogFileEvent write fOnLogFileEvent; {0.11}{0.29}
      property DoVerifyDomain : boolean read fDoVerifyDomain write fDoVerifyDomain;
      property DoDNSLists : boolean read fDoDNSLists write fDoDNSLists;
      property DoPOPAuth : boolean read fDoPOPAuth write fDoPOPAuth;
      property DoSpamTrapping : boolean read fDoSpamTrapping write fDoSpamTrapping;
      property MailWriteDir : shortstring read fMailWriteDir write fMailWriteDir;
      property LocalBlackListNotice : shortstring read fLocalBlackListNotice write fLocalBlackListNotice;
      property ForwardMail : boolean read fForwardMail write fForwardMail;
      property CheckUser : boolean read fCheckUser write fCheckUser; {0.33}
      property ExtendedErrorRecording : boolean read fExtendedErrorRecording write fExtendedErrorRecording; {0.35}
      property SpamTrapRejectTime : integer read fSpamTrapRejectTime write fSpamTrapRejectTime; {0.47}

      procedure AddLocalDomain(const Domain : string);
      procedure AddLocalNet(LocalNet : PSubNet);
      procedure AddToLocalBlackList(BlackListEntry : PSubNet);
      procedure AddToWhiteList(WhiteListEntry : PSubNet);
      procedure AddLocalUser(const UserName : string);
      procedure AddSpamTrapUser(const UserName : string);
      procedure AddDomainToSkipUserVerify(const DomainToSkip : string);
      procedure AddDNSBlockList(BlockListEntry : PDNSListRec);
    end;

implementation

uses
  Windows,SysUtils,FileCtrl,
  GlobalWinshoe,WinsockIntf,
  SMTPFUNC,DEFDATE;


//------------------------------------------------------------------------------
//                        Start RecipientList Code
//------------------------------------------------------------------------------

constructor TRecipientList.Create(AOwner : TComponent);
begin
  Inherited Create(AOwner);
  fList := TList.Create;
  fList.Clear;
end;

destructor TRecipientList.Destroy;
var
  Idx : integer;
  aPtr : pRecipientRec;
begin
  if fList.Count > 0 then begin
    For Idx := 0 To (fList.count -1) do begin
       aPtr := fList.Items[Idx];
       Dispose(aPtr);
     end;
   end;
  fList.Free;
  Inherited Destroy;
end;

procedure TRecipientList.Clear;
var
  aPtr : pRecipientRec;
  Idx : integer;
begin
  if fList.count > 0 then begin
    For Idx := 0 To (fList.count -1) do begin
      aPtr := fList.Items[Idx];
      Dispose(aPtr);
    end;
    fList.Clear;
  end;
end;

function TRecipientList.GetCount : integer;
begin
  GetCount := fList.Count;
end;

procedure TRecipientList.Add(RecipientItem : tRecipientRec);
var
  pRecipientItem : pRecipientRec;
begin
  New(pRecipientItem);
  pRecipientItem^ := RecipientItem;
  fList.Add(pRecipientItem);
end;

function TRecipientList.GetRecipientRec(Idx: integer):tRecipientRec;
var
  RecipientItem : tRecipientRec;
begin
  FillChar(RecipientItem,SizeOf(RecipientItem),0);
  Result := RecipientItem;
  if (Idx < fList.Count) And (Idx >= 0) then begin
    Result := tRecipientRec(fList.Items[Idx]^);
  end;
end;

//------------------------------------------------------------------------------
//                        end RecipientList Code
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
//                        start accesscontrol code
//------------------------------------------------------------------------------

constructor TAccessControl.Create(AOwner : TComponent);
begin
  Inherited Create(AOwner);
  LocalDomainsCritSection := TCriticalSection.Create;
  LocalUsersCritSection := TCriticalSection.Create;
  SpamTrapUsersCritSection := TCriticalSection.Create;
  SkipCritSection := TCriticalSection.Create;
  LocalNetCritSection := TCriticalSection.Create;
  LocalBlackListCritSection := TCriticalSection.Create;
  WhiteListCritSection := TCriticalSection.Create;
  DNSListCritSection := TCriticalSection.Create;
  SpamTrapListCritSection := TCriticalSection.Create;
  fDNSBlockLists := TList.Create;
  fDNSBlockLists.Clear;
  fSpamTrapRelayList := TList.Create;
  fSpamTrapRelayList.Clear;
  fLocalNets := TList.Create;
  fLocalNets.Clear;
  fLocalBlackList := TList.Create;
  fLocalBlackList.Clear;
  fWhiteList := TList.Create;
  fWhiteList.Clear;
  fLocalDomains := TStringList.Create;
  fLocalDomains.Clear;
  fLocalUsers := TStringList.Create;
  fLocalUsers.Clear;
  fSpamTrapUsers := TStringList.Create;
  fSpamTrapUsers.Clear;
  fSkipUserVerifyDomains := TStringList.Create;
  fSkipUserVerifyDomains.Clear;
end;

destructor TAccessControl.Destroy;
var
  Loop : longint;
begin
  for Loop := 0 to (fDNSBlockLists.Count-1) do begin
    dispose(fDNSBlockLists[Loop]);
    fDNSBlockLists[Loop] := nil;
  end;
  fDNSBlockLists.Pack;
  fDNSBlockLists.Free;
  for Loop := 0 to (fSpamTrapRelayList.Count-1) do begin
    dispose(fSpamTrapRelayList[Loop]);
    fSpamTrapRelayList[Loop] := nil;
  end;
  fSpamTrapRelayList.Pack;
  fSpamTrapRelayList.Free;
  for Loop := 0 to (fLocalNets.Count-1) do begin
    dispose(fLocalNets[Loop]);
    fLocalNets[Loop] := nil;
  end;
  fLocalNets.Pack;
  fLocalNets.Free;
  for Loop := 0 to (fLocalBlackList.Count-1) do begin
    dispose(fLocalBlackList[Loop]);
    fLocalBlackList[Loop] := nil;
  end;
  fLocalBlackList.Pack;
  fLocalBlackList.Free;
  for Loop := 0 to (fWhiteList.Count-1) do begin
    dispose(fWhiteList[Loop]);
    fWhiteList[Loop] := nil;
  end;
  fWhiteList.Pack;
  fWhiteList.Free;
  fLocalDomains.Free;
  fLocalUsers.Free;
  fSpamTrapUsers.Free;
  fSkipUserVerifyDomains.Free;
  LocalDomainsCritSection.Free;
  LocalUsersCritSection.Free;
  SpamTrapUsersCritSection.Free;
  SkipCritSection.Free;
  LocalNetCritSection.Free;
  LocalBlackListCritSection.Free;
  WhiteListCritSection.Free;
  DNSListCritSection.Free;
  SpamTrapListCritSection.Free;
  Inherited Destroy;
end;

function TAccessControl.IsInNetList(const IP : string; fNetList : TList): boolean;
var
  ActualIP : string;
  idx : integer;
  TempNet : PSubNet;
  IPAddy : longword;
begin
  IsInNetList := false;
  ActualIP := Trim(IP);
  IPAddy := ParseIPAddress(ActualIP);
  if fNetList.Count > 0 then begin
    For idx := 0 To fNetList.Count - 1 do begin
      TempNet := fNetList[Idx];
      IsInNetList := Result or ((IPAddy and TempNet.Mask) = TempNet.Net);
    end;
  end;
end;

function TAccessControl.IsLocalNet(const IP : string) : boolean;
begin
  LocalNetCritSection.Enter;
  try
    IsLocalNet := IsInNetList(IP,fLocalNets);
  finally
    LocalNetCritSection.Leave;
  end;
end;

function TAccessControl.IsWhiteListed(const IP : string) : boolean;
begin
  WhiteListCritSection.Enter;
  try
    IsWhiteListed := IsInNetList(IP,fWhiteList);
  finally
    WhiteListCritSection.Leave;
  end;
end;

function TAccessControl.IsBlackListedLocally(const IP : string) : boolean;
begin
  LocalBlackListCritSection.Enter;
  try;
    IsBlackListedLocally := IsInNetList(IP,fLocalBlackList);
  finally
    LocalBlackListCritSection.Leave;
  end;
end;

function TAccessControl.IsDNSBlockListed(const IP : string; var Reason : string) : boolean;
var
  BlockList : TDNSListRec;
  NumLists,Loop : integer;
  Reverse : string;
  IPos : integer;
  Query : string;
  S : string;
  aHost: PHostEnt;
  SendingIP : string;
begin
  IsDNSBlockListed := false;
  Reason := '';
  SendingIP := IP;
  Reverse := '';
  IPos := pos('.',SendingIP);
  while (IPos>0) do begin
    Reverse := '.'+copy(SendingIP,1,IPos-1)+Reverse;
    delete(SendingIP,1,IPos);
    IPos := pos('.',SendingIP);
  end;
  Reverse := SendingIP+Reverse;
  DNSListCritSection.Enter;
  try
    NumLists := fDNSBlockLists.Count;
  finally
    DNSListCritSection.Leave;
  end;
  for Loop := 0 to NumLists-1 do begin
    DNSListCritSection.Enter;
    try
      BlockList :=  PDNSListRec(fDNSBlockLists[Loop])^;
    finally
      DNSListCritSection.Leave;
    end;
    Query := Reverse + '.' + BlockList.DNSZoneFileName+#0+#0;
    aHost := WinsockInterface.GetHostByName(PChar(Query));
    if (aHost <> nil) then begin
      S := BlockList.DNSReponseString;
      IPos := pos('%s',S);
      if (IPos > 0) then begin
        delete(S,Ipos,2);
        insert(IP,S,IPos);
      end;
      Reason := S;
      IsDNSBlockListed := true;
      exit;
    end;
  end;
end;

function TAccessControl.IsLocalDomain(const Domain : string): boolean;
var
  ActualDomain : string;
  idx : integer;
begin
  IsLocalDomain := false;
  ActualDomain := Trim(UpperCase(Domain));
  LocalDomainsCritSection.Enter;
  try
    for idx := 0 To (fLocalDomains.Count - 1) do begin
      if CompareText(ActualDomain,Trim(UpperCase(fLocalDomains[Idx]))) = 0 then begin
        Result := True;
        break;
      end;
    end;
  finally
    LocalDomainsCritSection.Leave;
  end;
end;

function TAccessControl.IsLocalUser(const Name : string): boolean; {0.33}
var
  Domain,ActualName : string;
  P,idx : integer;
begin
  IsLocalUser := false;
  ActualName := UpperCase(Trim(Name));
  P := pos('@',ActualName);
  Domain := '';
  if (P > 0) then begin
    Domain := ActualName;
    delete(Domain,1,P);
    SkipCritSection.Enter;
    try
      for idx := 0 To (fSkipUserVerifyDomains.Count - 1) do begin {0.38}
        if (Domain = UpperCase(Trim(fSkipUserVerifyDomains[Idx]))) then begin
          IsLocalUser := true;
          break;
        end;
      end;
    finally
      SkipCritSection.Leave;
    end;
    if Result then exit;
    ActualName := copy(ActualName,1,P-1);
  end;
  LocalUsersCritSection.Enter;
  try
    for idx := 0 to (fLocalUsers.Count - 1) do begin
      IsLocalUser := Result or (ActualName = UpperCase(Trim(fLocalUsers.Strings[Idx])));
    end;
  finally
    LocalUsersCritSection.Leave;
  end;
end;

function TAccessControl.IsSpamTrapUser(const Name : string): boolean; {0.47}
var
  ActualName : string;
  P,idx : integer;
begin
  IsSpamTrapUser := false;
  ActualName := UpperCase(Trim(Name));
  P := pos('@',ActualName);
  if (P > 0) then
    ActualName := copy(ActualName,1,P-1);
  SpamTrapUsersCritSection.Enter;
  try
    for idx := 0 to (fSpamTrapUsers.Count - 1) do begin
      IsSpamTrapUser := Result or (ActualName = UpperCase(Trim(fSpamTrapUsers.Strings[Idx])));
    end;
  finally
    SpamTrapUsersCritSection.Leave;
  end;
end;

procedure TAccessControl.RegisterSpamRelay(const IP : string);
var
  IPAddy : longword;
  Rec : PSpamTrapRec;
begin
  IPAddy := ParseIPAddress(IP);
  new(Rec);
  with Rec^ do begin
    CreateTime := Now;
    Address := IPAddy;
  end;
  SpamTrapListCritSection.Enter;
  try
    fSpamTrapRelayList.Add(Rec);
  finally
    SpamTrapListCritSection.Leave;
  end;
end;

function TAccessControl.IsSpamTrapRelay(const IP : string; TimeOut : longint) : boolean;
var
  Loop : longint;
  IPAddy : longword;
  Rec : PSpamTrapRec;
begin
  IPAddy := ParseIPAddress(IP);
  IsSpamTrapRelay := false;
  SpamTrapListCritSection.Enter;
  try
    for Loop := 0 to fSpamTrapRelayList.Count-1 do begin
      Rec := fSpamTrapRelayList[Loop];
      if (((Now-Rec^.CreateTime)*MinutesPerDay) > (TimeOut)) then begin
        dispose(Rec);
        fSpamTrapRelayList[Loop] := nil;
      end else begin
        if (Rec^.Address = IPAddy) then begin
          IsSpamTrapRelay := true;
          break;
        end;
      end;
    end;
    fSpamTrapRelayList.Pack;
  finally
    SpamTrapListCritSection.Leave;
  end;
end;

procedure TAccessControl.AddLocalDomain(const Name : string);
begin
  LocalDomainsCritSection.Enter;
  try
    fLocalDomains.Add(Name);
  finally
    LocalDomainsCritSection.Leave;
  end;
end;

procedure TAccessControl.AddLocalUser(const Name : string);
begin
  LocalUsersCritSection.Enter;
  try
    fLocalUsers.Add(Name);
  finally
    LocalUsersCritSection.Leave;
  end;
end;

procedure TAccessControl.AddSpamTrapUser(const Name : string);
begin
  SpamTrapUsersCritSection.Enter;
  try
    fSpamTrapUsers.Add(Name);
  finally
    SpamTrapUsersCritSection.Leave;
  end;
end;

procedure TAccessControl.AddDomainToSkipUserVerify(const DomainToSkip : string);
begin
  SkipCritSection.Enter;
  try
    fSkipUserVerifyDomains.Add(DomainToSkip);
  finally
    SkipCritSection.Leave;
  end;
end;

procedure TAccessControl.AddLocalNet(LocalNet : PSubNet);
begin
  LocalNetCritSection.Enter;
  try
    fLocalNets.Add(LocalNet);
  finally
    LocalNetCritSection.Leave;
  end;
end;

procedure TAccessControl.AddToLocalBlackList(BlackListEntry : PSubNet);
begin
  LocalBlackListCritSection.Enter;
  try
    fLocalBlackList.Add(BlackListEntry);
  finally
    LocalBlackListCritSection.Leave;
  end;
end;

procedure TAccessControl.AddToWhiteList(WhiteListEntry : PSubNet);
begin
  WhiteListCritSection.Enter;
  try
    fWhiteList.Add(WhiteListEntry);
  finally
    WhiteListCritSection.Leave;
  end;
end;

procedure TAccessControl.AddDNSBlockList(BlockListEntry : PDNSListRec);
begin
  DNSListCritSection.Enter;
  try
    fDNSBlockLists.Add(BlockListEntry);
  finally
    DNSListCritSection.Leave;
  end;
end;

//------------------------------------------------------------------------------
//                        Start SMTP Message Code
//------------------------------------------------------------------------------
constructor TSMTPMsg.Create(AOwner: TComponent);
begin                           { Create }
  Inherited Create(AOwner);
  fRecipients := TRecipientList.Create(Self);
  fData := TSTringList.Create;
end;                            { Create }

destructor TSMTPMsg.Destroy;
begin                           { Destroy }
  Fdata.Free;
  FRecipients.Free;
  Inherited Destroy;
end;                            { Destroy }

procedure TSMTPMsg.Clear(FirstTime : boolean);
begin
  fRecipients.Clear;
  fData.Clear;
  fFromStr := '';
  fDateTime := 0;
  fMsgId := '';
  if FirstTime then begin
    fSendingDomain := '';
    fVerifiedSendingDomain := '';
    fSendingIP := '255.255.255.255';
  end;
end;

//------------------------------------------------------------------------------
//  Start of TWinshoeSMTPServer Code
//------------------------------------------------------------------------------

constructor TWinshoeSMTPServer.Create(AOwner: TComponent);
begin
  Inherited Create(AOwner);
  SetDefaultDateTimeStrings;
  fAccessControl := tAccessControl.Create(Self);
  Port :=  WSPORT_SMTP;;
  fServerName := GetLocalFQDNName; {0.32}
  if (fServerName = '') then {0.20}
    fServerName := LocalAddress;
end;

destructor TWinShoeSMTPServer.Destroy;
begin
  fAccessControl.Free;
  inherited Destroy;
end;

function TWinshoeSMTPServer.GetLocalFQDNName : string; {0.32}
var
  aHost: PHostEnt;
begin
  aHost := WinsockInterface.gethostbyname(PChar(LocalName));
  if aHost = nil then begin
    CheckForSocketError(SOCKET_ERROR);
    Result := '';
  end else begin
    Result := String(PChar(aHost.h_name));
  end;
end;

//------------------------------------------------------------------------------
//  Execute support procedures
//------------------------------------------------------------------------------

function TWinShoeSMTPServer.DoPOPLogVerify(SendingIP : string) : boolean;
var
  Local, Handled : boolean;
begin
  Result := false;
  if not DoPOPAuth then exit;
  if Assigned(OnPOPAuth) then OnPOPAuth(SendingIP, Local , Handled);
  if Handled then
    Result := Local;
end;

function TWinShoeSMTPServer.VerifyLocalSender(SendingIP : string) : boolean;
begin
  VerifyLocalSender := false;
  if fAccessControl.IsLocalNet(SendingIP) then begin
    VerifyLocalSender := true;
  end else begin
    VerifyLocalSender := DoPOPLogVerify(SendingIP);
  end;
end;

function TWinShoeSMTPServer.PassedBlackLists(SendingIP : string;
                                             var Failure : string) : boolean;
var
  Reason : string;
begin
  PassedBlackLists := true;
  Failure := '';
  if fAccessControl.IsWhiteListed(SendingIP) then exit; {0.41}
  if fAccessControl.IsBlackListedLocally(SendingIP) or
     fAccessControl.IsSpamTrapRelay(SendingIP,SpamTrapRejectTime) then begin {0.41}{0.47}
    Failure := fLocalBlackListNotice;
    PassedBlackLists := false;
    exit;
  end;
  if not fDoDNSLists then exit;
  if fAccessControl.IsDNSBlockListed(SendingIP,Reason) then begin
    PassedBlackLists := false;
    Failure := Reason;
  end;
end;

function TWinShoeSMTPServer.VerifySendingDomain(SendingIP : string) : string;
begin
  VerifySendingDomain := 'unverified';
  if DoVerifyDomain then
    VerifySendingDomain := Trim(TWinShoe.ResolveIP(SendingIP));
  if (Result = '') then {0.19}
    VerifySendingDomain := 'unverified';
end;

//------------------------------------------------------------------------------
//  onconnect procedure thread handler
//------------------------------------------------------------------------------

procedure TWinShoeSMTPServer.DoConnect(Thread: TWinshoeServerThread);
begin                           { DoConnect }
  Thread.Connection.WriteLn('220 '+fServerName +' '+ProductName+' SMTP Receiver Version '+VersionString+' Ready');
end;                            { DoConnect }

//------------------------------------------------------------------------------
//  main thread handler
//------------------------------------------------------------------------------

function TWinshoeSMTPServer.DoExecute(Thread: TWinshoeServerThread): boolean;
var
  RcvdStr,  ArgStr,  sCmd: string;
  Loop,SMTPCommand : TSMTPCommands;
  SessionMsg : tSMTPMsg;

begin                           { DoExecute }
  Result := inherited DoExecute(Thread);
  if Result then exit;
  SessionMsg := tSMTPMsg.Create(Self);
  try { 0.45 }
    SessionMsg.Clear(true);
    with SessionMsg do begin
      SendingIP := Thread.Connection.PeerAddress;
      {do local IP or POPAuth first}
      fLocalSender := VerifyLocalSender(SendingIP);
      fFailReason := '';
      {do DNS blacklists only if not a local sender}
      if fLocalSender or PassedBlackLists(SendingIP,fFailReason) then begin
        fDenyAllCommands := false;
      end else begin
        fDenyAllCommands := true;
        if Assigned(OnLogFileEvent) then OnLogFileEvent(LogFileDateTime+' '+SendingIP+' 554 access denied, '+fFailReason);{0.11}{0.29}
      end;
      VerifiedSendingDomain := VerifySendingDomain(SendingIP);
    end;
    with Thread.Connection do begin
    while Connected do begin
      RcvdStr := ReadLn;
      ArgStr := RcvdStr;
      sCmd := UpperCase(GetFirstTokenDeleteFromArg(ArgStr,CHAR32));
      SMTPCommand := cError;
      for Loop := FirstSMTPCommand to LastSMTPCommand do begin
        if (sCmd = UpperCase(SMTPCommands[Loop])) then
          SMTPCommand := Loop;
      end;
      case SMTPCommand of
        cHELO : DoCommandHELO(Thread, ArgStr, SessionMsg);
        cMAIL : DoCommandMAIL(Thread, ArgStr, SessionMsg);
        cRCPT : DoCommandRCPT(Thread, ArgStr, SessionMsg);
        cDATA : DoCommandDATA(Thread, SessionMsg);
        cSEND : DoCommandSEND(Thread, ArgStr, SessionMsg);
        cSOML : DoCommandSOML(Thread, ArgStr, SessionMsg);
        cSAML : DoCommandSAML(Thread, ArgStr, SessionMsg);
        cRSET : DoCommandRSET(Thread, SessionMsg);
        CVRFY : DoCommandVRFY(Thread, ArgStr);
        cEXPN : DoCommandEXPN(Thread, ArgStr);
        cHELP : DoCommandHELP(Thread, ArgStr);
        cNOOP : DoCommandNOOP(Thread);
        cQuit : DoCommandQUIT(Thread);
        cTURN : DoCommandTURN(Thread, SessionMsg);
        else DoCommandError(Thread, RcvdStr);
      end; {case}
    end; {while}
    end; {with}
  finally { 0.45 }
    SessionMsg.Free;
  end;
end;                            { DoExecute }

//------------------------------------------------------------------------------
//  dummy SMTP commands
//------------------------------------------------------------------------------

procedure TWinShoeSMTPServer.DoCommandVRFY(Thread: TWinshoeServerThread;
                                           aName: string);
begin                           { DoCommandVRFY }
  Thread.Connection.WriteLn('250 OK '+aname);
end;                            { DoCommandVRFY }

procedure TWinShoeSMTPServer.DoCommandNOOP(Thread: TWinshoeServerThread);
begin                           { DoCommandNOOP }
  Thread.Connection.WriteLn('250 NOOP Performed');
end;                            { DoCommandNOOP }

procedure TWinShoeSMTPServer.DoCommandEXPN(Thread: TWinshoeServerThread;
                                            aListName: string);
begin                           { DoCommandEXPN }
  Thread.Connection.WriteLn('550 EXPN Access Denied');
end;                            { DoCommandEXPN }

procedure TWinShoeSMTPServer.DoCommandSEND(Thread: TWinshoeServerThread;FromStr: string;
                                            SessionMsg: tSMTPMsg);
begin                           { DoCommandSEND }
  Thread.Connection.WriteLn('502 SEND messaging prohibited.');
end;                            { DoCommandSEND }

procedure TWinShoeSMTPServer.DoCommandSOML(Thread: TWinshoeServerThread; FromStr: string;
                                            SessionMsg: tSMTPMsg);
begin                           { DoCommandSOML }
  Thread.Connection.WriteLn('502 SOML messaging prohibited.');
end;                            { DoCommandSOML }

procedure TWinShoeSMTPServer.DoCommandSAML(Thread: TWinshoeServerThread; FromStr: string;
                                            SessionMsg: tSMTPMsg);
begin                           { DoCommandSAML }
  Thread.Connection.WriteLn('502 SAML messaging prohibited.');
end;                            { DoCommandSAML }

procedure TWinShoeSMTPServer.DoCommandHELP(Thread: TWinshoeServerThread;
                                            aCommand: string);
begin                           { DoCommandHELP }
  Thread.Connection.WriteLn('502 HELP no help available.');
end;                            { DoCommandHELP }

procedure TWinShoeSMTPServer.DoCommandTURN(Thread: TWinshoeServerThread;
                                            SessionMsg: tSMTPMsg);
begin                           { DoCommandTURN }
  Thread.Connection.WriteLn('502 TURN prohibited.');
end;                            { DoCommandTURN }

procedure TWinShoeSMTPServer.DoCommandError(Thread: TWinshoeServerThread;
                                             ArgStr: string);
begin                           { DoCommandError }
  Thread.Connection.WriteLn('500 Syntax Error, Command Unrecognized '+ArgStr);
end;                            { DoCommandError }

//------------------------------------------------------------------------------
//  live SMTP commands
//------------------------------------------------------------------------------

procedure TWinShoeSMTPServer.DoCommandHELO(Thread: TWinshoeServerThread; Domain: string;
                                           SessionMsg: tSMTPMsg);
var
  S : string;
begin                           { DoCommandHELO }
  with SessionMsg do begin
    if (length(SendingDomain) > 0) then begin
      Thread.Connection.WriteLn('503 HELO Command is out of sequence');
    end else begin
      SendingDomain := Domain;
      S := '';
      if SessionMsg.fDenyAllCommands then
        S := ', access restricted';
      Thread.Connection.WriteLn('250 OK '+Domain+' ('+VerifiedSendingDomain+' ['+SendingIP+'])'+S);
    end;
  end;
end;                            { DoCommandHELO }

procedure TWinShoeSMTPServer.DoCommandMAIL(Thread: TWinshoeServerThread; FromStr: string;
                                           SessionMsg: tSMTPMsg);
var
  Domain,PrettyAddress,MachineAddress : string;
begin                           { DoCommandMAIL }
  if ParseAddress(FromStr,Domain,PrettyAddress,MachineAddress) then begin
    SessionMsg.FromStr := MachineAddress;
    Thread.Connection.WriteLn('250 OK '+PrettyAddress);
  end else begin
    StripToFrom(FromStr);
    SessionMsg.FromStr := FromStr;
    Thread.Connection.WriteLn('250 OK '+FromStr);
{  RFC 2505 violation v0.18
    Thread.Connection.WriteLn('501 Bad From Address '+ FromStr);
}
  end;
end;                            { DoCommandMAIL }

procedure TWinShoeSMTPServer.DoCommandRCPT(Thread: TWinshoeServerThread; ToStr: string;
                                           SessionMsg: tSMTPMsg);
var
  aRecipient : TRecipientRec;
  Domain,PrettyAddress,MachineAddress : string;
begin                           { DoCommandRCPT }
  if ParseAddress(ToStr,Domain,PrettyAddress,MachineAddress) then begin
    fillchar(aRecipient,sizeof(aRecipient),0);
    with aRecipient do begin
      ValidatedAddress := MachineAddress;
      LocalDomain := fAccessControl.IsLocalDomain(Domain);
    end;
    if (not SessionMsg.fDenyAllCommands) then begin
      if DoSpamTrapping and fAccessControl.IsSpamTrapUser(MachineAddress) then begin
        SessionMsg.fDenyAllCommands := true;
        fAccessControl.RegisterSpamRelay(SessionMsg.SendingIP);
        SessionMsg.Recipients.Clear;
        SessionMsg.fFailReason := LocalBlackListNotice;
        Thread.Connection.WriteLn('550 <'+MachineAddress+'>... user not valid');
        if fExtendedErrorRecording and Assigned(OnLogFileEvent) then
          OnLogFileEvent(LogFileDateTime+' '+SessionMsg.SendingIP+' 550 <'+MachineAddress+'> spam trap address hit');{0.47}
      end else if aRecipient.LocalDomain then begin
        if ((not fCheckUser) or fAccessControl.IsLocalUser(MachineAddress)) then begin {0.33}
          Thread.Connection.WriteLn('250 OK '+PrettyAddress);
          SessionMsg.Recipients.Add(aRecipient);
        end else begin
          Thread.Connection.WriteLn('550 <'+MachineAddress+'>... user not valid');
          if fExtendedErrorRecording and Assigned(OnLogFileEvent) then
            OnLogFileEvent(LogFileDateTime+' '+SessionMsg.SendingIP+' 550 <'+MachineAddress+'> user not valid');{0.35}
        end;
      end else if SessionMsg.fLocalSender or fForwardMail then begin {0.21}
        Thread.Connection.WriteLn('250 OK '+PrettyAddress);
        SessionMsg.Recipients.Add(aRecipient);
      end else begin
        Thread.Connection.WriteLn('550 <'+MachineAddress+'>... Relaying denied');
        if fExtendedErrorRecording and Assigned(OnLogFileEvent) then
          OnLogFileEvent(LogFileDateTime+' '+SessionMsg.SendingIP+' 550 <'+MachineAddress+'> relaying denied');{0.35}
      end
    end else begin
      Thread.Connection.WriteLn('554 access denied, '+SessionMsg.fFailReason);
    end;
  end else begin
    Thread.Connection.WriteLn('501 Syntax Error in RCPT '+TosTr);
  end;
end;                            { DoCommandRCPT }

procedure TWinShoeSMTPServer.DoCommandDATA(Thread: TWinshoeServerThread;
                                           SessionMsg: tSMTPMsg);
var
  RcvdStr : string;
  Finished : boolean;
begin                           { DoCommandData }
  with Thread.Connection do begin
    if SessionMsg.fDenyAllCommands then begin {0.26}
      Thread.Connection.WriteLn('554 access denied, '+SessionMsg.fFailReason);
    end else if (SessionMsg.Recipients.Count < 1) then begin
      Thread.Connection.WriteLn('503 No valid recipients specified');
    end else begin
      WriteLn('354 Ready for data');
      // (!!0.01} Modification to handle  extra Period and blank line bug
      // Modification by Roman Evstyunichev... thanks.
      Finished := false;
      repeat
        RcvdStr := ReadLn;
       {v0.13-0.15 undoes !!0.01}
       {mod here is for SMTPDS compatibility... it doesn't remove it. MVM 013101}
        Finished := ((Length(RcvdStr) = 1) and (RcvdStr[1] = '.'));
        SessionMsg.fData.add(RcvdStr);
      until Finished;
      if WriteMail(Thread, SessionMsg) then begin
        WriteLn('250 Message received OK');
      end else begin
        WriteLn('451 ERR Requested Action Aborted');
      end;
      SessionMsg.Clear(false); {0.31}
    end;
  end;
end;                            { DoCommandData }

procedure TWinShoeSMTPServer.DoCommandRSET(Thread: TWinshoeServerThread;
                                            SessionMsg: tSMTPMsg);
begin                           { DoCommandRSET }
  SessionMsg.Clear(false);
  Thread.Connection.WriteLn('250 OK Reset Completed');
end;                            { DoCommandRSET }

procedure TWinShoeSMTPServer.DoCommandQUIT(Thread: TWinshoeServerThread);
begin                          { DoCommandQUIT }
  try
    Thread.Connection.WriteLn('221 '+fServerName
                            +' closing');
  finally
    Thread.Connection.Disconnect;
  end;
end;                            { DoCommandQUIT }

//------------------------------------------------------------------------------
//  SMTP command support procedures
//------------------------------------------------------------------------------

function TWinShoeSMTPServer.GetNextID : longint;
var
  ID : longint;
  Handled : boolean;
begin
  if Assigned(OnNextIDEvent) then OnNextIDEvent(ID , Handled);
  if not Handled then
    ID := 1;
  GetNextID := ID;
end;

function TWinShoeSMTPServer.WriteMessageToFile(SessionMsg: TSMTPMsg): boolean;
var
  Recp : tRecipientRec;
  FiName ,S,aStr : string;
  f: Text;
  mIdx : integer;
  DateFound,IDFound : boolean;
  LogFileEntry : string;
begin
  Result := true;
  LogFileEntry := '';
  with SessionMsg do begin
    if (SessionMsg.Recipients.Count < 1) then begin
      Result := false;
      exit;
    end;
    DateFound := false;
    IDFound := false;
    for mIdx := 0 To (fData.count - 1) do begin
      S := UpperCase(Trim(Fdata.Strings[mIdx]));
      if (pos('DATE',S) = 1) then DateFound := true;
      if (pos('RECEIVED',S) = 1) then DateFound := true; {0.24}
      if (pos('MESSAGE-ID',S) = 1) then IDFound := true;
      if (S = '') then break; {end of header}
    end;
    try
      try
        FiName := FixDirBackSlash(fMailWriteDir)+ChangeFileExt(MsgID,'.MSG');
        {fdata is of type tstringlist, use windows call to write to file 0.22}
        AssignFile(f,FiName);
        ReWrite(F);
        Astr := 'Received: from '+ SendingDomain +' ('+ VerifiedSendingDomain+
                ' ['+SendingIP+']) by '+fServerName;
        fData.Insert(0,Astr);
        Astr := ' ('+ProductName+' '+ VersionString + ') with SMTP id <'+MsgId+'@'+fServerName+'>;';
        fData.Insert(1,Astr);
        Astr := ' '+GetInternetDateTimeStr(Now);
        fData.Insert(2,Astr);
        if not IDFound then begin
          S := 'Message-ID: <'+MsgId+'@'+fServerName+'>';
          fData.Insert(3,S);
        end;
        if not DateFound then begin
          Astr := 'Date:'+AStr;
          fData.Insert(3,Astr);
        end;
{        fData.SaveToFile(FiName);}
        For mIdx := 0 To (fData.count - 1) do begin
          WriteLn(F,Fdata.Strings[mIdx]);
        end;
      except
        on Exception do
          Result := false;
        end;
    finally
      CloseFile(f);
    end;
    try
      try
        LogFileEntry := '<'+MsgID+'@'+fServerName+'> '+
                        LogFileDateTime+' '+SendingIP+' '+
                        SendingDomain+' '+FromStr;{0.11}
        FiName := FixDirBackSlash(fMailWriteDir)+ChangeFileExt(MsgID,'.RCP');
        AssignFile(f,FiName);
        ReWrite(F);
        Astr := 'Message-ID: <'+MsgID+'@'+fServerName+'>';
        WriteLn(F,Astr);
        Astr := 'Return-path: '+FromStr;
        WriteLn(F,Astr);
        For mIdx := 0 To Recipients.count - 1 do begin
          Recp := Recipients.GetRecipientRec(mIdx);
          Astr := 'Recipient: '+Recp.ValidatedAddress;
          LogFileEntry := LogFileEntry + ' '+Recp.ValidatedAddress;
          WriteLn(F,Astr);
        end;
        if Assigned(OnLogFileEvent) then OnLogFileEvent(LogFileEntry);{0.11}
      Except
        On Exception do
          Result := false;
        end;
    finally
      CloseFile(f);
    end;
  end;
end;

function TWinShoeSMTPServer.WriteMail(Thread: TWinshoeServerThread;
                                      SessionMsg:tSMTPMsg) : boolean;
var
  S : string;
begin
  S := IntToStr(GetNextID);
  while (length(S) < 10) do
    S := '0' + S;
  SessionMsg.MsgID := 'B'+S;
  WriteMail := WriteMessageToFile(SessionMsg);
end;

//------------------------------------------------------------------------------
//  access control assign procedures
//------------------------------------------------------------------------------

procedure TWinshoeSMTPServer.AddLocalDomain(const Domain : string);
begin
  fAccessControl.AddLocalDomain(Domain);
end;

procedure TWinshoeSMTPServer.AddLocalUser(const UserName : string);
begin
  fAccessControl.AddLocalUser(UserName);
end;

procedure TWinshoeSMTPServer.AddSpamTrapUser(const UserName : string);
begin
  fAccessControl.AddSpamTrapUser(UserName);
end;

procedure TWinshoeSMTPServer.AddDomainToSkipUserVerify(const DomainToSkip : string);
begin
  fAccessControl.AddDomainToSkipUserVerify(DomainToSkip);
end;

procedure TWinshoeSMTPServer.AddLocalNet(LocalNet : PSubNet);
begin
  fAccessControl.AddLocalNet(LocalNet);
end;

procedure TWinshoeSMTPServer.AddToLocalBlackList(BlackListEntry : PSubNet);
begin
  fAccessControl.AddToLocalBlackList(BlackListEntry);
end;

procedure TWinshoeSMTPServer.AddToWhiteList(WhiteListEntry : PSubNet);
begin
  fAccessControl.AddToWhiteList(WhiteListEntry);
end;

procedure TWinshoeSMTPServer.AddDNSBlockList(BlockListEntry : PDNSListRec);
begin
  fAccessControl.AddDNSBlockList(BlockListEntry);
end;

end.
